# python3-opsi
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
# and published under the Terms of the GNU Affero General Public License.
# Text of the AGPL: http://www.gnu.org/licenses/agpl-3.0-standalone.html
# credits: http://www.opsi.org/credits/

#***************************************************************************
# Subversion:
# $Revision: 40 $
# $Author: oertel $
# $Date: 2015-01-16 18:17:00 +0100 (Fr, 16 Jan 2015) $
#***************************************************************************

#

try:
    import ConfigParser
except:
    import configparser
from OPSI.Util.File import IniFile
from OPSI.Util.File import TextFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from OPSI.Util import blowfishDecrypt
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass
try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse
from os.path import splitext, basename
import platform
import re, xml.dom.minidom, os, gzip
import crypt
import codecs
from OpenSSL.crypto import FILETYPE_PEM, load_certificate
from base64 import b64encode
from os import urandom

if not bootimageVersion or (int(bootimageVersion) < 20210618):
    raise Exception(u"This product requires opsi-linux-bootimage version >= 20210618")
try:
    if not backend.backend_info()['opsiVersion'].startswith('4.'):
        raise ValueError('No opsi 4.x backend present')
except Exception:
    raise Exception(u"This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

if isOpsi43:
    show_message(u"Installing Product: %s" % (productId), 10)

opsiVersion = backend.backend_info()['opsiVersion']

# Do hardware inventory
logger.notice(u"Fetching opsi hw audit configuration")
hwconfig = backend.auditHardware_getConfig()
if isOpsi43:
    show_message(u"Running hardware inventory", 10)
logger.notice(u"Running hardware inventory")
auditHardwareOnHosts = auditHardware(config = hwconfig, hostId = clientId)

logger.notice(u"Sending hardware information to service")
backend.auditHardwareOnHost_setObsolete(clientId)
backend.auditHardwareOnHost_createObjects(auditHardwareOnHosts)

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsipreparelib.py','/tmp')
exec(open('/tmp/opsipreparelib.py').read())

initOpsiSetupLib()

useGpt = False
inUefi = inUefiMode()
inSecureboot = False
if inUefi:
    try:
        inSecureboot = isSecureboot()
    except NameError:
        isSecureboot = False

    if inSecureboot:
        logger.notice(u"we are running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are running in Secure Boot mode")
    else:
        logger.notice(u"we are not running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are not running in Secure Boot mode")
    useGpt = True

if inUefi:
    useGpt = True
    currentbootentry = uefiGetActBootEntry()
    if currentbootentry:
        uefiNetbootLabel = currentbootentry['label']
        logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
        backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label")
        backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
    else:
        logger.notice('No usable current (net) boot entry found')


# create network path
if (depot._url.startswith('smb:') or depot._url.startswith('cifs:')):
    match = re.search('^smb://([^/]+)/([^/]+)(.*)$', depot._url, re.IGNORECASE)
    if not match:
        raise Exception("Bad depot-URL '%s'" % depot._url)
    hn = match.group(1)
    sn = match.group(2)
    pn = match.group(3)

depotDir = '/var/lib/opsi/depot'

proxy = productProperties.get('proxy','')

install_opsi_client_agent = False
do_opsi_client = "false"
if (str(productProperties.get('install_opsi-client-agent', 'true')).lower() == 'true'):
    install_opsi_client_agent = True
    do_opsi_client = "true"

install_unattended = True
if (str(productProperties.get('install_unattended', 'true')).lower() == 'false'):
    install_unattended = False

installation_method = productProperties.get('installation_method', 'kexec')
if inSecureboot:
    installation_method = 'reboot'
    logger.notice("We are running in SecureBoot mode, enforcing reboot installation_method")

language = productProperties.get('language','de_DE')
keymap = productProperties.get('console_keymap', 'german')

partition_disk = productProperties.get('partition_disk', 'first')

my_broadcast = getBroadcastAddress(ipaddress, netmask)
my_network = getNetworkAddress(ipaddress, netmask)

logger.notice(u"Guess network interface name....")
# in 13.1 the network interfaces are renamed: http://www.freedesktop.org/wiki/Software/systemd/PredictableNetworkInterfaceNames/
# so we have get the networkinterface name from ifconfig

networkdevice = ''
logger.notice(u"We are on a predictable network interface names bootimage")
for device, lease in leasesFiles.items():
    networkdevice = device


logger.notice(u"Will configure network interface name: %s" % networkdevice)

# Get and encrypt passwords
# user
userpassclear = productProperties.get('user_password', 'linux123')
userpassenc = shadow_password_hash(userpassclear)

# root
rootpassclear = productProperties.get('root_password', 'linux123')
rootpassenc = shadow_password_hash(rootpassclear)

if (proxy == ''):
    do_proxy = 'true'
    proxy_addr = proxy
else:
    do_proxy = 'false'
    proxy_addr = ''

# get machine architecture
myarchitecture = getArchitecture()
if myarchitecture == 'amd64':
    scriptMessageSubject.setMessage(u"Installing 64 Bit system....")
    logger.notice(u"Installing 64 Bit system....")
    myarchitectureBits = '64'
else:
    scriptMessageSubject.setMessage("Installing 32 Bit system....")
    logger.notice(u"Installing 32 Bit system....")
    myarchitectureBits = '32'

if isOpsi43:
    show_message("Preparing disk", 20)
scriptMessageSubject.setMessage("Create temporary place on disk.... ")
logger.notice(u"Create temporary place on disk.... ")
#prepareDisk()
target = '/mnt/hd'
systemPartitionNumber = 1
if inUefi:
    systemPartitionNumber = 3 
    if installation_method == 'reboot':
        grubPartitionNumber = 3 
        systemPartitionNumber +=1 
        grubPartitionSizeM = 2000
        grubPartitionLabel = 'INSTALLER'
systemPartitionSizeM = 2000
systemPartitionLabel  = 'LINUX'
blockAlignment = True

# Preprepare disks
logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
lvmClearVolumeGroups()
try:
    execute("cryptsetup remove lukslvm")
except:
    pass
# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
# use first disk
disk = disks[0]
if partition_disk == "first":
    part_disk = disk.device
else:
    part_disk = partition_disk

if useGpt:
    logger.notice(u"Clear partition table of %s" % (disk.device))
    execute('/usr/sbin/sgdisk --zap-all %s' % (disk.device))
    execute('/usr/sbin/sgdisk --clear -g -p %s' % (disk.device))
    if inUefi:
        deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
        deleteAllEntriesFromEFIBootmgr('Windows Boot Manager')
        deleteAllEntriesFromEFIBootmgr('ubuntu')
        deleteAllEntriesFromEFIBootmgr('debian')
        deleteAllEntriesFromEFIBootmgr('installer')
        deleteAllEntriesFromEFIBootmgr('opsibootimage')
        deleteAllEntriesFromEFIBootmgr('CentOS')
        deleteAllEntriesFromEFIBootmgr('Red Hat')
        logger.notice(u"Create EFI system partition on %s1" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 1::100M -t 1:ef00 -c 1:\"EFI system partition\" -p %s' % (disk.device))
        execute('/bin/sleep 2')
        execute('/usr/sbin/sgdisk -p %s' % (disk.device))
        logger.notice(u"Make file system on EFI system partition on %s1" % (disk.device))
        normalDiskDevices = ['/dev/sda', '/dev/vda', '/dev/sdb']
        if any(normalDisk in disk.device for normalDisk in normalDiskDevices):
            execute('/sbin/mkfs.vfat %s1' % (disk.device))
        else:
            execute('/sbin/mkfs.vfat %sp1' % (disk.device))
        logger.notice(u"Create MS reserved partition on %s2" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 2::+128M -t 2:0c01 -c 2:\"MS reserved\" -p %s' % (disk.device))
else:
    disk.deletePartitionTable()

depotAddress = depotUrl.split('/')[2].replace(':4447','')
if installation_method == 'reboot':
    if inUefi:
        grubTarget = '/mnt/grub/'
        checkCreateDir(grubTarget)
        execute('/usr/sbin/sgdisk --info=%d %s' % (grubPartitionNumber, disk.device))
        createPartitionEx("installer", grubPartitionNumber, grubPartitionLabel, 0, grubPartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "63" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "fat32", gptFileSystem="0700")
        execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 --attributes %d:clear:0 -p %s' % (grubPartitionNumber, grubPartitionNumber, grubPartitionNumber, disk.device))
        testPartitionEx(disk.device, grubPartitionNumber, useGpt)
        checkClearPartition("none", disk.device, grubPartitionNumber)
        createFilesystemEx("bootimage", grubPartitionNumber, "bootimage", "fat32", disk.device , useGpt)
        checkMount(grubTarget, disk.device, grubPartitionNumber, "vfat")
        logger.notice("Copying files to partition %d" % grubPartitionNumber)
        depot.copy('/%s/installfiles64/*' % productId, grubTarget)

        customisationLine = f'keymap={keymap} install=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/isocontent/ splash=silent add_efi_memmap gpt acpi_rsdp=EFI'
        newLines = []
        with open('%s/grub/grub.cfg' % (grubTarget), 'r+') as grubConfig:
            for line in grubConfig:
                if line.strip().startswith('linux'):
                    newLines.append('%s %s\n' % (line.rstrip(), customisationLine.rstrip()))
                    continue
                newLines.append(line)

        with open('%s/grub/grub.cfg' % (grubTarget), 'w') as grubConfig:
            grubConfig.writelines(newLines)
        execute(f'sed -i s_inst.ks=file:/ks.cfg__g {grubTarget}/grub/grub.cfg')
        execute(f'sed -i s/vmlinuz/linux/g {grubTarget}/grub/grub.cfg')

    else:
        logger.warning(u"reboot installation method only works using UEFI. Reverting back to kexec method")
        installation_method = "kexec"

fdiskexitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
createPartitionEx("system", systemPartitionNumber, systemPartitionLabel, 0, systemPartitionSizeM, gptAutoAppend = True, boot = False, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ext4", gptFileSystem="8300")
createFilesystemEx("system", systemPartitionNumber, systemPartitionLabel, "ext4", disk.device, useGpt)
systemPartitionDevicePath = getPartitionDevicePath(disk.device, systemPartitionNumber)

# Partition table deleted => delete all netboot product states
deletePocs = []
for poc in backend.productOnClient_getObjects(clientId = clientId):
    if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
       ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
        deletePocs.append(poc)
backend.productOnClient_deleteObjects(deletePocs)

diskSizeM = (int(disk.size)/(1024*1024))
mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
swapSize = (int(mem_bytes)/(1024*1024))
rootSize = diskSizeM-swapSize-150

diskSizeM = str(diskSizeM) + ' MB'
rootSize = str(rootSize) + ' MB'
swapSize = str(swapSize) + ' MB'

if isOpsi43:
    show_message("Mounting partitions and copying data", 40)
# Mount temporary partition
checkCreateDir(target)
checkClearMountpoint(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkClearPartition(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkMount(target, disk.device, systemPartitionNumber, "ext4", systemPartitionDevicePath)
checkCreateDir(target+'/tmp')

scriptMessageSubject.setMessage("Getting kernel and initrd.... ")
logger.notice(u"Getting kernel and initrd.... ")
try:
    depot.copy(SCRIPT_PATH + '/isocontent/boot/x86_64/loader/linux', target+'/tmp')
    depot.copy(SCRIPT_PATH + '/isocontent/boot/x86_64/loader/initrd', target+'/tmp')
except Exception as fnferror:
    raise Exception("isocontent directory doesn't seem to be filled")

scriptMessageSubject.setMessage("Uncompressing initrd.... ")
logger.notice(u"Uncompressing initrd.... ")
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("mv %s/tmp/initrd %s/tmp/initrd.xz" % (target,target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
    execute("xz -d %s/tmp/initrd.xz"% (target))
except:
    pass
if os.path.exists("%s/tmp/initrd.xz"% (target)):
    execute("mv %s/tmp/initrd.xz %s/tmp/initrd.gz" % (target,target))
    try:
        execute("ls -l %s/tmp/initrd*" % (target))
        execute("gunzip %s/tmp/initrd.gz"% (target))
    except:
        pass

scriptMessageSubject.setMessage("Patching initrd.... ")
logger.notice(u"Mounting initrd ....")
execute("mkdir -p %s/tmp/mnt" % (target))
execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd" % (target))

logger.notice(u"Copy profile for patch....")
if inUefi:
    depot.copy(SCRIPT_PATH+'/'+productId+'-efi.xml', target+'/tmp/installation_profile')
    #depot.copy(SCRIPT_PATH+'/control.xml', target+'/tmp/installation_profile')
else:
    depot.copy(SCRIPT_PATH+'/'+productId+'.xml', target+'/tmp/installation_profile')
if os.path.exists(DEPOT_MOUNT + '/' + SCRIPT_PATH+'/custom/'+productId+'.xml'):
    depot.copy(SCRIPT_PATH+'/custom/'+productId+'.xml', target+'/tmp/installation_profile')


execute(f"wget --no-check-certificate -P /tmp/ https://{fqdn}:{pckey}@{depotAddress}:4447/ssl/opsi-ca-cert.pem")
execute(f"cat /tmp/opsi-ca-cert.pem >> {target}/tmp/mnt/var/lib/ca-certificates/ca-bundle.pem")
execute(f"cp /tmp/opsi-ca-cert.pem {target}/tmp/mnt/var/lib/ca-certificates/pem/")
cert_file = "/tmp/opsi-ca-cert.pem"
with codecs.open(cert_file, "r", "utf-8") as file:
    c = load_certificate(FILETYPE_PEM, file.read().encode("ascii"))
    hex_hash = f"{c.subject_name_hash():x}"
if hex_hash:
    execute(f"cd {target}/tmp/mnt/var/lib/ca-certificates/pem/ && ln -s opsi-ca-cert.pem {hex_hash}.0")
    execute(f"ls -hal {target}/tmp/mnt/var/lib/ca-certificates/pem/{hex_hash}.0")

if isOpsi43:
    show_message("patching files", 90)
logger.notice(u"Prepare patch input....")
with open(PATCHA_IN, "a") as myfile:
    myfile.write('networkaddress=%s\n' % my_network)
    myfile.write('broadcastaddress=%s\n' % my_broadcast)
    myfile.write('nameserveraddress=%s\n' % nameserver)
    myfile.write('do_opsi_client=%s\n' % do_opsi_client)
    myfile.write('userpassenc=%s\n' % userpassenc)
    myfile.write('rootpassenc=%s\n' % rootpassenc)
    myfile.write('do_proxy=%s\n' % do_proxy)
    myfile.write('proxy_addr=%s\n' % proxy_addr)
    myfile.write('part_disk=%s\n' % part_disk)
    myfile.write('networkdevice=%s\n' % networkdevice)
    myfile.write('service_ip_address=%s\n' % (nextserver.replace("/rpc", "").rstrip("/")))
    myfile.write('depotAddress=%s\n' % (depotUrl.split('/')[2].replace(':4447','')))
    myfile.write('client_id=%s\n' % (fqdn))
    myfile.write('client_key=%s\n' % (pckey))
if inUefi:
    with open(PATCHA_IN, "a") as myfile:
        myfile.write('disk=%s\n' % disk.device)
        myfile.write('rootsize=%s\n' % rootSize)
        myfile.write('swapsize=%s\n' % swapSize)

logger.notice(u"Here comes the PATCHA_IN (%s):" % PATCHA_IN)
execute('cat %s' % (PATCHA_IN))

logger.notice(u"Patching installation_profile")
if install_opsi_client_agent == False:
    lines = []
    with open(target + '/tmp/installation_profile', 'r') as inputxml:
        lines = inputxml.readlines()
    outlines = []
    keep = True
    for line in lines:
        if "mkdir -p /var/tmp/opsi-linux-client-agent" in line:
            keep = False
        if keep:
            outlines.append(line)
        if "systemctl enable olca-install.service" in line:
            keep = True
    with open(target + '/tmp/installation_profile', 'w') as outputxml:
        outputxml.writelines(outlines)

execute('/usr/local/bin/patcha %s' % (target + '/tmp/installation_profile'))
execute('/usr/local/bin/patcha -f %s %s' % (PATCHA_IN, target + '/tmp/installation_profile'))

logger.notice(u"Copy opsinetmount file to initrd...")
depot.copy(SCRIPT_PATH+'/opsinetmount', target+'/tmp/mnt/sbin')
execute("chmod 777 %s/tmp/mnt/sbin/opsinetmount" % (target))

if install_unattended:
    logger.notice(u"Copy profile file to initrd...")
    copy(target+'/tmp/installation_profile', target+'/tmp/mnt/autoinst.xml')
    logger.notice(u"Here comes the installation profile:")
    execute('cat %s/tmp/mnt/autoinst.xml' % (target))

logger.notice(u"Copy opsi-client-agent file to initrd...")
execute('mkdir -p %s/tmp/mnt/opsitmp/' % (target))
copy(PATCHA_IN, target+'/tmp/mnt/opsitmp/')

scriptMessageSubject.setMessage("Rebuild patched initrd.... ")
logger.notice(u"Rebuild patched initrd.... ")
execute("rm %s/tmp/initrd" % (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("cd %s/tmp/mnt ; find . | cpio -o -H newc > %s/tmp/initrd" % (target, target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
if installation_method == 'reboot':
    execute('/bin/efibootmgr -c --disk %s --part %s -L installer --loader \\\\bootx64.efi' % (disk.device, grubPartitionNumber))
    setFirstUefiBoot('installer')
    setNextUefiBoot('installer')
    execute('/bin/efibootmgr -v')
    execute("cp %s/tmp/initrd* %s" % (target, grubTarget))
    execute("cp %s/tmp/linux %s" % (target, grubTarget))
    depot.copy(f"{SCRIPT_PATH}/isocontent/EFI/BOOT/bootx64.efi", grubTarget)
    depot.copy(f"{SCRIPT_PATH}/isocontent/EFI/BOOT/grub.efi", grubTarget)
    execute(f"cp {grubTarget}/grub/grub.cfg {grubTarget}")
    execute("cp %s/tmp/mnt/autoinst.xml %s" % (target, grubTarget))
else:
    scriptMessageSubject.setMessage("Prepare loading kernel.... ")
    logger.notice(u"Prepare loading kernel ....")
    kernel_file = target+'/tmp/linux'
    #append_line = "splash=silent pci=biosirq"
    #append_line = "splash=silent liveinstall"
    if inUefi:
        append_line = 'splash=silent add_efi_memmap gpt acpi_rsdp=EFI'
    else:
        append_line = 'splash=silent'
    inird_file = target+'/tmp/initrd'
    #execute("%s --load %s  --append='%s' --initrd=%s" % (which('kexec'), kernel_file, append_line, inird_file))
    execute(f"{which('kexec')} --load {kernel_file}  --append='install=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/isocontent/ {append_line}' --initrd={inird_file}")

logger.notice(u"handle_setupAfterInstall ....")
handle_setupAfterInstall()

# Set product installation status
productOnClient.setInstallationStatus('unknown')
#productOnClient.setActionProgress('installing')
productOnClient.setActionResult('')
productOnClient.setLastAction('setup')

try:
    # Write log to service
    logger.notice(u"Writing log to service")

    log = codecs.open(LOG_FILE, 'r', 'utf-8', 'replace')
    data = log.read()
    log.close()

    logger.setFileLevel(LOG_ERROR)
    if backend:
        try:
            try:
                ui.getMessageBox().addText(_(u"Writing log to service\n"))
            except:
                pass
            backend.log_write(logType = 'bootimage', data = data, objectId = clientId, append = False)
        except Exception as e:
            logger.error(u"Failed to write log to service: %s" % e)

except Exception as e:
    logger.logException(e)

if installation_method == 'reboot':
    scriptMessageSubject.setMessage("Rebooting machine ")
    logger.notice(u"Rebooting machine")
    reboot()
else:
    scriptMessageSubject.setMessage("Execute new kernel.... ")
    logger.notice(u"Execute new kernel ....")
    execute("%s -e" % (which('kexec')))
